@@ -26,9 +26,6 @@
 #include "gossip-utils.h"
 #include "gossip-contact.h"
 
-#define DEBUG_MSG(x) 
-/* #define DEBUG_MSG(args) g_printerr args ; g_printerr ("\n"); */
-
 #define GET_PRIV(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), GOSSIP_TYPE_CONTACT, GossipContactPriv))
 
 /* Make sure any additions to the private contact structure are copied
@@ -217,28 +214,18 @@ contact_finalize (GObject *object)
 	g_free (priv->avatar);
 
 	if (priv->presences) {
-		GList *l;
-
-		for (l = priv->presences; l; l = l->next) {
-			g_object_unref (l->data);
-		}
-
+		g_list_foreach (priv->presences, (GFunc) g_object_unref, NULL);
 		g_list_free (priv->presences);
 	}
 
 	if (priv->groups) {
-		GList *l;
-
-		for (l = priv->groups; l; l = l->next) {
-			g_free (l->data);
-		}
-
+		g_list_foreach (priv->groups, (GFunc) g_free, NULL);
 		g_list_free (priv->groups);
 	}
 	
 	g_object_unref (priv->account);
 	
-	(* G_OBJECT_CLASS (parent_class)->finalize) (object);
+	(G_OBJECT_CLASS (parent_class)->finalize) (object);
 }
 
 static void
@@ -557,6 +544,8 @@ gossip_contact_set_id (GossipContact *contact,
 
 	g_free (priv->id);
 	priv->id = g_strdup (id);
+
+	g_object_notify (G_OBJECT (contact), "id");
 }
 
 void
@@ -572,6 +561,8 @@ gossip_contact_set_name (GossipContact *contact,
 
 	g_free (priv->name);
 	priv->name = g_strdup (name);
+
+	g_object_notify (G_OBJECT (contact), "name"); 
 }
 
 void 
@@ -594,6 +585,9 @@ gossip_contact_set_avatar (GossipContact *contact,
 		priv->avatar = NULL;
 		priv->avatar_size = 0;
 	}
+
+	g_object_notify (G_OBJECT (contact), "avatar"); 
+	g_object_notify (G_OBJECT (contact), "avatar-size"); 
 }
 
 void
@@ -616,6 +610,8 @@ gossip_contact_set_account (GossipContact *contact,
 	} else {
 		priv->account = NULL;
 	}
+
+	g_object_notify (G_OBJECT (contact), "account"); 
 }
 
 void 
@@ -689,21 +685,22 @@ gossip_contact_set_groups (GossipContact *contact,
 			   GList         *groups)
 {
 	GossipContactPriv *priv;
-	GList             *l;
+	GList             *old_groups, *l;
 
 	g_return_val_if_fail (GOSSIP_IS_CONTACT (contact), FALSE);
 	
 	priv = GET_PRIV (contact);
 
-	if (priv->groups) {
-		g_list_foreach (priv->groups, (GFunc)g_free, NULL);
-		priv->groups = NULL;
-	}
+	old_groups = priv->groups;
+	priv->groups = NULL;
 
 	for (l = groups; l; l = l->next) {
 		priv->groups = g_list_append (priv->groups, 
 					      g_strdup (l->data));
 	}
+	
+	g_list_foreach (old_groups, (GFunc) g_free, NULL);
+	g_list_free (old_groups);
 
 	return TRUE;
 }
@@ -719,10 +716,10 @@ gossip_contact_set_subscription (GossipContact      *contact,
 	priv = GET_PRIV (contact);
 
 	priv->subscription = subscription;
+
+	g_object_notify (G_OBJECT (contact), "subscription");
 }
 
-
-
 gint
 gossip_contact_compare (gconstpointer a,
 			gconstpointer b)
@@ -736,7 +733,7 @@ gossip_contact_compare (gconstpointer a,
 	id_a = gossip_contact_get_id (GOSSIP_CONTACT (a));
 	id_b = gossip_contact_get_id (GOSSIP_CONTACT (b));
 
-	/* FIXME: Maybe use utf8 strcmp? */
+	/* FIXME: Maybe use utf8 strcmp? At least jabber ids are UTF8. */
 	return strcmp (id_a, id_b);
 }
 
@@ -747,10 +744,6 @@ gossip_contact_name_compare (gconstpointer a,
 	g_return_val_if_fail (GOSSIP_IS_CONTACT (a), 0);
 	g_return_val_if_fail (GOSSIP_IS_CONTACT (b), 0);
 	
-	DEBUG_MSG (("Contact: Comparing %s vs. %s",
-		    gossip_contact_get_name (GOSSIP_CONTACT (a)),
-		    gossip_contact_get_name (GOSSIP_CONTACT (b))));
-
 	return strcmp (gossip_contact_get_name (GOSSIP_CONTACT (a)),
 		       gossip_contact_get_name (GOSSIP_CONTACT (b)));
 }
