@@ -19,17 +19,16 @@
  */
 
 #include <config.h>
+#include <string.h>
 
 #include "libgossip-marshal.h"
-
+#include "gossip-debug.h"
 #include "gossip-session.h"
 
-#define DEBUG_MSG(x) 
-/* #define DEBUG_MSG(args) g_printerr args ; g_printerr ("\n"); */
+#define DEBUG_DOMAIN "Session"
 
 #define GET_PRIV(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), GOSSIP_TYPE_SESSION, GossipSessionPriv))
 
-
 typedef struct _GossipSessionPriv  GossipSessionPriv;
 
 struct _GossipSessionPriv {
@@ -43,35 +42,39 @@ struct _GossipSessionPriv {
 	GList                *contacts;
 
 	guint                 connected_counter;
+	guint                 connecting_counter;
 
 	GHashTable           *timers; /* connected time */
 };
 
-
 typedef struct {
-	gchar         *contact_id;
+	const gchar   *contact_id;
 	GossipAccount *account;
 } FindAccount;
 
-
-struct CountAccounts {
+typedef struct {
 	GossipSession *session;
 	guint          connected;
+	guint          connecting;
 	guint          disconnected;
-};
-
+} CountAccounts;
 
 typedef struct {
 	GossipSession *session;
 	GList         *accounts;
 } GetAccounts;
 
-
 typedef struct {
 	GossipSession *session;
+	gboolean       connect_all;
 	gboolean       startup;
-} ConnectAccounts;
+} ConnectData;
 
+typedef struct {
+	gboolean       first;
+	gboolean       last;
+	gboolean       same;
+} ConnectFind;
 
 static void            gossip_session_class_init                 (GossipSessionClass   *klass);
 static void            gossip_session_init                       (GossipSession        *session);
@@ -83,6 +86,7 @@ static void            session_protocol_logged_in                (GossipProtocol
 								  GossipSession        *session);
 static void            session_protocol_logged_out               (GossipProtocol       *protocol,
 								  GossipAccount        *account,
+								  gint                  reason,
 								  GossipSession        *session);
 static void            session_protocol_new_message              (GossipProtocol       *protocol,
 								  GossipMessage        *message,
@@ -99,7 +103,7 @@ static void            session_protocol_contact_presence_updated (GossipProtocol
 static void            session_protocol_contact_removed          (GossipProtocol       *protocol,
 								  GossipContact        *contact,
 								  GossipSession        *session);
-static void            session_protocol_composing_event          (GossipProtocol       *protocol,
+static void            session_protocol_composing                (GossipProtocol       *protocol,
 								  GossipContact        *contact,
 								  gboolean              composing,
 								  GossipSession        *session);
@@ -112,15 +116,6 @@ static void            session_protocol_error                    (GossipProtocol
 								  GossipSession        *session);
 static GossipProtocol *session_get_protocol                      (GossipSession        *session,
 								  GossipContact        *contact);
-static void            session_get_accounts_foreach_cb           (GossipAccount        *account,
-								  GossipProtocol       *protocol,
-								  GetAccounts          *data);
-static void            session_count_accounts_foreach_cb         (GossipAccount        *account,
-								  GossipProtocol       *protocol,
-								  struct CountAccounts *ca);
-static void            session_find_account_foreach_cb           (GossipAccount        *account,
-								  GossipProtocol       *protocol,
-								  FindAccount          *fa);
 static void            session_account_added_cb                  (GossipAccountManager *manager,
 								  GossipAccount        *account,
 								  GossipSession        *session);
@@ -129,17 +124,9 @@ static void            session_account_removed_cb                (GossipAccountM
 								  GossipSession        *session);
 static void            session_connect                           (GossipSession        *session,
 								  GossipAccount        *account);
-static void            session_connect_foreach_cb                (GossipAccount        *account,
-								  GossipProtocol       *protocol,
-								  ConnectAccounts      *data);
 static void            session_disconnect                        (GossipSession        *session,
 								  GossipAccount        *account);
-static void            session_disconnect_foreach_cb             (GossipAccount        *account,
-								  GossipProtocol       *protocol,
-								  GossipSession        *session);
 	
-
-/* signals */
 enum {
 	CONNECTING,
 	CONNECTED,
@@ -156,21 +143,17 @@ enum {
 	CONTACT_UPDATED,
 	CONTACT_PRESENCE_UPDATED,
 	CONTACT_REMOVED,
-	COMPOSING_EVENT,
+	COMPOSING,
 
 	/* Used for protocols to retreive information from UI */
 	GET_PASSWORD,
 	LAST_SIGNAL
 };
 
+static guint    signals[LAST_SIGNAL] = {0};
 
 G_DEFINE_TYPE (GossipSession, gossip_session, G_TYPE_OBJECT);
 
-
-static guint    signals[LAST_SIGNAL] = {0};
-static gpointer parent_class;
-
-
 static void
 gossip_session_class_init (GossipSessionClass *klass)
 {
@@ -178,8 +161,6 @@ gossip_session_class_init (GossipSessionClass *klass)
 	
 	object_class = G_OBJECT_CLASS (klass);
 
-	parent_class = g_type_class_peek_parent (klass);
-
 	object_class->finalize = session_finalize;
 
 	signals[CONNECTING] = 
@@ -224,9 +205,9 @@ gossip_session_class_init (GossipSessionClass *klass)
 			      G_SIGNAL_RUN_LAST,
 			      0,
 			      NULL, NULL,
-			      libgossip_marshal_VOID__POINTER,
+			      libgossip_marshal_VOID__OBJECT_OBJECT,
 			      G_TYPE_NONE, 
-			      1, G_TYPE_POINTER);
+			      2, GOSSIP_TYPE_ACCOUNT, GOSSIP_TYPE_PROTOCOL);
 
 	signals[PROTOCOL_CONNECTED] = 
 		g_signal_new ("protocol-connected",
@@ -234,9 +215,9 @@ gossip_session_class_init (GossipSessionClass *klass)
 			      G_SIGNAL_RUN_LAST,
 			      0,
 			      NULL, NULL,
-			      libgossip_marshal_VOID__POINTER_POINTER,
+			      libgossip_marshal_VOID__OBJECT_OBJECT,
 			      G_TYPE_NONE, 
-			      2, G_TYPE_POINTER, G_TYPE_POINTER);
+			      2, GOSSIP_TYPE_ACCOUNT, GOSSIP_TYPE_PROTOCOL);
 	
 	signals[PROTOCOL_DISCONNECTED] = 
 		g_signal_new ("protocol-disconnected",
@@ -244,9 +225,9 @@ gossip_session_class_init (GossipSessionClass *klass)
 			      G_SIGNAL_RUN_LAST,
 			      0,
 			      NULL, NULL,
-			      libgossip_marshal_VOID__POINTER_POINTER,
+			      libgossip_marshal_VOID__OBJECT_OBJECT_INT,
 			      G_TYPE_NONE, 
-			      2, G_TYPE_POINTER, G_TYPE_POINTER);
+			      3, GOSSIP_TYPE_ACCOUNT, GOSSIP_TYPE_PROTOCOL, G_TYPE_INT);
 
 	signals[PROTOCOL_DISCONNECTING] = 
 		g_signal_new ("protocol-disconnecting",
@@ -254,9 +235,9 @@ gossip_session_class_init (GossipSessionClass *klass)
 			      G_SIGNAL_RUN_LAST,
 			      0,
 			      NULL, NULL,
-			      libgossip_marshal_VOID__POINTER,
+			      libgossip_marshal_VOID__OBJECT_OBJECT,
 			      G_TYPE_NONE, 
-			      1, G_TYPE_POINTER);
+			      2, GOSSIP_TYPE_ACCOUNT, GOSSIP_TYPE_PROTOCOL);
 	
 	signals[PROTOCOL_ERROR] = 
 		g_signal_new ("protocol-error",
@@ -264,9 +245,9 @@ gossip_session_class_init (GossipSessionClass *klass)
 			      G_SIGNAL_RUN_LAST,
 			      0,
 			      NULL, NULL,
-			      libgossip_marshal_VOID__POINTER_POINTER_POINTER,
+			      libgossip_marshal_VOID__OBJECT_OBJECT_POINTER,
 			      G_TYPE_NONE, 
-			      3, G_TYPE_POINTER, G_TYPE_POINTER, G_TYPE_POINTER);
+			      3, GOSSIP_TYPE_PROTOCOL, GOSSIP_TYPE_ACCOUNT, G_TYPE_POINTER);
 
 	signals[NEW_MESSAGE] = 
 		g_signal_new ("new-message",
@@ -274,9 +255,9 @@ gossip_session_class_init (GossipSessionClass *klass)
 			      G_SIGNAL_RUN_LAST,
 			      0,
 			      NULL, NULL,
-			      libgossip_marshal_VOID__POINTER,
+			      libgossip_marshal_VOID__OBJECT,
 			      G_TYPE_NONE,
-			      1, G_TYPE_POINTER);
+			      1, GOSSIP_TYPE_MESSAGE);
 
 	signals[PRESENCE_CHANGED] = 
 		g_signal_new ("presence-changed",
@@ -284,9 +265,9 @@ gossip_session_class_init (GossipSessionClass *klass)
 			      G_SIGNAL_RUN_LAST,
 			      0,
 			      NULL, NULL,
-			      libgossip_marshal_VOID__POINTER,
+			      libgossip_marshal_VOID__OBJECT,
 			      G_TYPE_NONE,
-			      1, G_TYPE_POINTER);
+			      1, GOSSIP_TYPE_PRESENCE);
 
 	signals[CONTACT_ADDED] = 
 		g_signal_new ("contact-added",
@@ -294,18 +275,18 @@ gossip_session_class_init (GossipSessionClass *klass)
 			      G_SIGNAL_RUN_LAST,
 			      0,
 			      NULL, NULL,
-			      libgossip_marshal_VOID__POINTER,
+			      libgossip_marshal_VOID__OBJECT,
 			      G_TYPE_NONE,
-			      1, G_TYPE_POINTER);
+			      1, GOSSIP_TYPE_CONTACT);
 	signals[CONTACT_UPDATED] = 
 		g_signal_new ("contact-updated",
 			      G_TYPE_FROM_CLASS (klass),
 			      G_SIGNAL_RUN_LAST,
 			      0,
 			      NULL, NULL,
-			      libgossip_marshal_VOID__POINTER,
+			      libgossip_marshal_VOID__OBJECT,
 			      G_TYPE_NONE,
-			      1, G_TYPE_POINTER);
+			      1, GOSSIP_TYPE_CONTACT);
 
 	signals[CONTACT_PRESENCE_UPDATED] = 
 		g_signal_new ("contact-presence-updated",
@@ -313,9 +294,9 @@ gossip_session_class_init (GossipSessionClass *klass)
 			      G_SIGNAL_RUN_LAST,
 			      0,
 			      NULL, NULL,
-			      libgossip_marshal_VOID__POINTER,
+			      libgossip_marshal_VOID__OBJECT,
 			      G_TYPE_NONE,
-			      1, G_TYPE_POINTER);
+			      1, GOSSIP_TYPE_CONTACT);
 
 	signals[CONTACT_REMOVED] = 
 		g_signal_new ("contact-removed",
@@ -323,27 +304,27 @@ gossip_session_class_init (GossipSessionClass *klass)
 			      G_SIGNAL_RUN_LAST,
 			      0,
 			      NULL, NULL,
-			      libgossip_marshal_VOID__POINTER,
+			      libgossip_marshal_VOID__OBJECT,
 			      G_TYPE_NONE,
-			      1, G_TYPE_POINTER);
-	signals[COMPOSING_EVENT] =
-		g_signal_new ("composing-event",
+			      1, GOSSIP_TYPE_CONTACT);
+	signals[COMPOSING] =
+		g_signal_new ("composing",
 			      G_TYPE_FROM_CLASS (klass),
 			      G_SIGNAL_RUN_LAST,
 			      0,
 			      NULL, NULL,
-			      libgossip_marshal_VOID__POINTER_BOOLEAN,
+			      libgossip_marshal_VOID__OBJECT_BOOLEAN,
 			      G_TYPE_NONE,
-			      2, G_TYPE_POINTER, G_TYPE_BOOLEAN);
+			      2, GOSSIP_TYPE_CONTACT, G_TYPE_BOOLEAN);
 	signals[GET_PASSWORD] =
 		g_signal_new ("get-password",
 			      G_TYPE_FROM_CLASS (klass),
 			      G_SIGNAL_RUN_LAST,
 			      0,
 			      NULL, NULL,
-			      libgossip_marshal_STRING__POINTER,
+			      libgossip_marshal_STRING__OBJECT,
 			      G_TYPE_STRING,
-			      1, G_TYPE_POINTER);
+			      1, GOSSIP_TYPE_ACCOUNT);
 	
 	g_type_class_add_private (object_class, sizeof (GossipSessionPriv));
 }
@@ -375,46 +356,34 @@ static void
 session_finalize (GObject *object)
 {
 	GossipSessionPriv *priv;
-	GList             *l;
 
 	priv = GET_PRIV (object);
 	
-	if (priv->accounts) {
-		g_hash_table_destroy (priv->accounts);
+	g_hash_table_destroy (priv->accounts);
+
+	g_list_foreach (priv->protocols, (GFunc) g_object_unref, NULL);
+	g_list_free (priv->protocols);
+
+	g_list_foreach (priv->contacts, (GFunc) g_object_unref, NULL);
+	g_list_free (priv->contacts);
+
+	if (priv->presence) {
+		g_object_unref (priv->presence);
 	}
 
-	if (priv->protocols) {
-		g_list_foreach (priv->protocols, 
-				(GFunc)g_object_unref,
-				NULL);
-		g_list_free (priv->protocols);
-	}
-
-	if (priv->contacts) {
-		for (l = priv->contacts; l; l = l->next) {
-			g_object_unref (l->data);
-		}
-
-		g_list_free (priv->contacts);
-	}
-
-	if (priv->timers) {
-		g_hash_table_destroy (priv->timers);
-	}
+	g_hash_table_destroy (priv->timers);
 
 	g_signal_handlers_disconnect_by_func (priv->account_manager,
 					      session_account_added_cb, 
-					      GOSSIP_SESSION (object));
+					      object);
 
 	g_signal_handlers_disconnect_by_func (priv->account_manager,
 					      session_account_removed_cb, 
-					      GOSSIP_SESSION (object));
+					      object);
 
- 	if (priv->account_manager) { 
- 		g_object_unref (priv->account_manager); 
- 	} 
+	g_object_unref (priv->account_manager); 
 
-	(* G_OBJECT_CLASS (parent_class)->finalize) (object);
+	(G_OBJECT_CLASS (gossip_session_parent_class)->finalize) (object);
 }
 
 static void
@@ -442,8 +411,8 @@ session_protocol_signals_setup (GossipSession  *session,
 	g_signal_connect (protocol, "contact-removed",
 			  G_CALLBACK (session_protocol_contact_removed),
 			  session);
-	g_signal_connect (protocol, "composing-event",
-			  G_CALLBACK (session_protocol_composing_event),
+	g_signal_connect (protocol, "composing",
+			  G_CALLBACK (session_protocol_composing),
 			  session);
 	g_signal_connect (protocol, "get-password",
 			  G_CALLBACK (session_protocol_get_password),
@@ -459,23 +428,22 @@ session_protocol_logged_in (GossipProtocol *protocol,
 			    GossipSession  *session)
 {
 	GossipSessionPriv *priv;
-	GTimer            *timer;
 
-	DEBUG_MSG (("Session: Protocol logged in"));
+	gossip_debug (DEBUG_DOMAIN, "Protocol logged in");
 
 	priv = GET_PRIV (session);
 
-	/* setup timer */
-	timer = g_timer_new ();
-	g_timer_start (timer);
-
 	g_hash_table_insert (priv->timers, 
 			     g_object_ref (account), 
-			     timer);
+			     g_timer_new ());
 	
-	/* update some status? */
+	/* Update some status? */
 	priv->connected_counter++;
 
+	if (priv->connecting_counter > 0) {
+		priv->connecting_counter--;
+	}
+
 	g_signal_emit (session, signals[PROTOCOL_CONNECTED], 0, account, protocol);
 
 	if (priv->connected_counter == 1) {
@@ -487,29 +455,39 @@ session_protocol_logged_in (GossipProtocol *protocol,
 static void
 session_protocol_logged_out (GossipProtocol *protocol, 
 			     GossipAccount  *account,
+			     gint            reason,
 			     GossipSession  *session) 
 {
 	GossipSessionPriv *priv;
 	gdouble            seconds;
 
 	seconds = gossip_session_get_connected_time (session, account);
-	DEBUG_MSG (("Session: Protocol logged out (after %.2f seconds)", seconds));
+	gossip_debug (DEBUG_DOMAIN, "Protocol logged out (after %.2f seconds)", 
+		      seconds);
 
 	priv = GET_PRIV (session);
 
-	/* remove timer */
+	/* Remove timer */
 	g_hash_table_remove (priv->timers, account);
 
-	/* update some status? */
+	/* Update some status? */
 	if (priv->connected_counter < 0) {
 		g_warning ("We have some issues in the connection counting");
 		return;
+	} 
+
+	/* Don't go lower than 0 */
+	if (priv->connected_counter > 0) {
+		priv->connected_counter--;
 	}
 
-	priv->connected_counter--;
-	
-	g_signal_emit (session, signals[PROTOCOL_DISCONNECTED], 0, account, protocol);
-	
+	if (priv->connecting_counter > 0) {
+		priv->connecting_counter--;
+	}
+
+	g_signal_emit (session, signals[PROTOCOL_DISCONNECTED], 0,
+		       account, protocol, reason);
+
 	if (priv->connected_counter == 0) {
 		/* Last connected protocol was disconnected */
 		g_signal_emit (session, signals[DISCONNECTED], 0);
@@ -532,8 +510,8 @@ session_protocol_contact_added (GossipProtocol *protocol,
 {
 	GossipSessionPriv *priv;
 
-	DEBUG_MSG (("Session: Contact added '%s'",
-		   gossip_contact_get_name (contact)));
+	gossip_debug (DEBUG_DOMAIN, "Contact added '%s'",
+		      gossip_contact_get_name (contact));
 
 	priv = GET_PRIV (session);
 	
@@ -548,8 +526,8 @@ session_protocol_contact_updated (GossipProtocol *protocol,
 				  GossipContact  *contact,
 				  GossipSession  *session)
 {
-	DEBUG_MSG (("Session: Contact updated '%s'",
-		   gossip_contact_get_name (contact)));
+	gossip_debug (DEBUG_DOMAIN, "Contact updated '%s'",
+		      gossip_contact_get_name (contact));
 
 	g_signal_emit (session, signals[CONTACT_UPDATED], 0, contact);
 }
@@ -559,8 +537,8 @@ session_protocol_contact_presence_updated (GossipProtocol *protocol,
 					   GossipContact  *contact,
 					   GossipSession  *session)
 {
-	DEBUG_MSG (("Session: Contact presence updated '%s'",
-		   gossip_contact_get_name (contact)));
+	gossip_debug (DEBUG_DOMAIN, "Contact presence updated '%s'",
+		      gossip_contact_get_name (contact));
 	g_signal_emit (session, signals[CONTACT_PRESENCE_UPDATED], 0, contact);
 }
 
@@ -571,8 +549,8 @@ session_protocol_contact_removed (GossipProtocol *protocol,
 {
 	GossipSessionPriv *priv;
 	
-	DEBUG_MSG (("Session: Contact removed '%s'",
-		   gossip_contact_get_name (contact)));
+	gossip_debug (DEBUG_DOMAIN, "Contact removed '%s'",
+		   gossip_contact_get_name (contact));
 
 	priv = GET_PRIV (session);
 	
@@ -583,20 +561,20 @@ session_protocol_contact_removed (GossipProtocol *protocol,
 }
 
 static void 
-session_protocol_composing_event (GossipProtocol *protocol,
-				  GossipContact  *contact,
-				  gboolean        composing,
-				  GossipSession  *session)
+session_protocol_composing (GossipProtocol *protocol,
+			    GossipContact  *contact,
+			    gboolean        composing,
+			    GossipSession  *session)
 {
 	GossipSessionPriv *priv;
 	
-	DEBUG_MSG (("Session: Contact %s composing:'%s'",
-		   composing ? "is" : "is not",
-		   gossip_contact_get_name (contact)));
+	gossip_debug (DEBUG_DOMAIN, "Contact %s composing:'%s'",
+		      composing ? "is" : "is not",
+		      gossip_contact_get_name (contact));
 
 	priv = GET_PRIV (session);
 	
-	g_signal_emit (session, signals[COMPOSING_EVENT], 0, contact, composing);
+	g_signal_emit (session, signals[COMPOSING], 0, contact, composing);
 }
 
 static gchar *
@@ -604,9 +582,9 @@ session_protocol_get_password (GossipProtocol *protocol,
 			       GossipAccount  *account,
 			       GossipSession  *session)
 {
-	gchar *password;
+	gchar *password = NULL;
 
-	DEBUG_MSG (("Session: Get password"));
+	gossip_debug (DEBUG_DOMAIN, "Get password");
 
 	g_signal_emit (session, signals[GET_PASSWORD], 0, account, &password);
 	
@@ -619,9 +597,16 @@ session_protocol_error (GossipProtocol *protocol,
 			GError         *error,
 			GossipSession  *session)
 {
+	GossipSessionPriv *priv;
 
-	DEBUG_MSG (("Session: Error:%d->'%s'", 
-		   error->code, error->message));
+	gossip_debug (DEBUG_DOMAIN, "Error:%d->'%s'", 
+		      error->code, error->message);
+
+	priv = GET_PRIV (session);
+
+	if (priv->connecting_counter > 0) {
+		priv->connecting_counter--;
+	}
 
 	g_signal_emit (session, signals[PROTOCOL_ERROR], 0, protocol, account, error); 
 }
@@ -691,6 +676,7 @@ gossip_session_new (GossipAccountManager *manager)
 		gossip_session_add_account (session, account);
 	}
 
+	g_list_foreach (accounts, (GFunc) g_object_unref, NULL);
 	g_list_free (accounts);
 
 	return session;
@@ -722,44 +708,41 @@ gossip_session_get_account_manager (GossipSession *session)
 	return priv->account_manager;
 }
 
+static void
+session_get_accounts_foreach_cb (GossipAccount  *account,
+				 GossipProtocol *protocol,
+				 GetAccounts    *ga)
+{
+	GossipSessionPriv *priv;
+
+	priv = GET_PRIV (ga->session);
+
+	ga->accounts = g_list_append (ga->accounts, g_object_ref (account));
+}
+
 GList *
 gossip_session_get_accounts (GossipSession *session)
 {
 	GossipSessionPriv *priv;
-	GetAccounts       *data;
+	GetAccounts        ga;
 	GList             *list;
 
 	g_return_val_if_fail (GOSSIP_IS_SESSION (session), NULL);
 	
 	priv = GET_PRIV (session);
 
-	data = g_new0 (GetAccounts, 1);
-	data->session = g_object_ref (session);
+	ga.session = session;
+	ga.accounts = NULL;
 
 	g_hash_table_foreach (priv->accounts, 
-			      (GHFunc)session_get_accounts_foreach_cb,
-			      data);
+			      (GHFunc) session_get_accounts_foreach_cb,
+			      &ga);
 
-	list = data->accounts;
-
-	g_object_unref (data->session);
-	g_free (data);
+	list = ga.accounts;
 
 	return list;
 }
 
-static void
-session_get_accounts_foreach_cb (GossipAccount   *account,
-				 GossipProtocol  *protocol,
-				 GetAccounts     *data)
-{
-	GossipSessionPriv *priv;
-
-	priv = GET_PRIV (data->session);
-
-	data->accounts = g_list_append (data->accounts, g_object_ref (account));
-}
-
 gdouble
 gossip_session_get_connected_time (GossipSession *session,
 				   GossipAccount *account)
@@ -781,40 +764,10 @@ gossip_session_get_connected_time (GossipSession *session,
 	return g_timer_elapsed (timer, &ms);
 }
 
-void 
-gossip_session_count_accounts (GossipSession *session,
-			       guint         *connected,
-			       guint         *disconnected)
-{
-	GossipSessionPriv    *priv;
-	struct CountAccounts  ca;
-	
-	g_return_if_fail (GOSSIP_IS_SESSION (session));
-	
-	priv = GET_PRIV (session);
-
-	ca.session = session;
-
-	ca.connected = 0;
-	ca.disconnected = 0;
-
-	g_hash_table_foreach (priv->accounts, 
-			      (GHFunc)session_count_accounts_foreach_cb,
-			      &ca);
-
-	if (connected) {
-		*connected = ca.connected;
-	}
-
-	if (disconnected) {
-		*disconnected = ca.disconnected;
-	}
-}
-
 static void
-session_count_accounts_foreach_cb (GossipAccount        *account,
-				   GossipProtocol       *protocol,
-				   struct CountAccounts *ca)
+session_count_accounts_foreach_cb (GossipAccount  *account,
+				   GossipProtocol *protocol,
+				   CountAccounts  *ca)
 {
 	GossipSessionPriv *priv;
 
@@ -824,6 +777,52 @@ session_count_accounts_foreach_cb (GossipAccount        *account,
 		ca->connected++;
 	} else {
 		ca->disconnected++;
+
+		if (gossip_protocol_is_connecting (protocol)) {
+			ca->connecting++;
+		}
+	}
+}
+
+void 
+gossip_session_count_accounts (GossipSession *session,
+			       guint         *connected,
+			       guint         *connecting,
+			       guint         *disconnected)
+{
+	GossipSessionPriv *priv;
+	CountAccounts      ca;
+	
+	g_return_if_fail (GOSSIP_IS_SESSION (session));
+	
+	priv = GET_PRIV (session);
+
+	ca.session = session;
+
+	ca.connected = 0;
+	ca.connecting = 0;
+	ca.disconnected = 0;
+
+	g_hash_table_foreach (priv->accounts, 
+			      (GHFunc) session_count_accounts_foreach_cb,
+			      &ca);
+
+	/* Here we update our own numbers, should we be doing this? I am not
+	 * sure but let's do it anyway :)
+	 */
+	priv->connected_counter = ca.connected;
+	priv->connecting_counter = ca.connecting;
+
+	if (connected) {
+		*connected = priv->connected_counter;
+	}
+
+	if (connecting) {
+		*connecting = priv->connecting_counter;
+	}
+
+	if (disconnected) {
+		*disconnected = ca.disconnected;
 	}
 }
 
@@ -840,11 +839,9 @@ gossip_session_add_account (GossipSession *session,
 
 	priv = GET_PRIV (session);
 
-	/* check this account is not already set up */
 	protocol = g_hash_table_lookup (priv->accounts, account);
-	
 	if (protocol) {
-		/* already added */
+		/* Already added. */
 		return TRUE;
 	}
 
@@ -852,19 +849,15 @@ gossip_session_add_account (GossipSession *session,
 	protocol = gossip_protocol_new_from_account_type (type);
 	g_return_val_if_fail (GOSSIP_IS_PROTOCOL (protocol), FALSE);
 
-	/* add to list */
 	priv->protocols = g_list_append (priv->protocols, 
 					 g_object_ref (protocol));
 
-	/* add to hash table */
 	g_hash_table_insert (priv->accounts, 
 			     g_object_ref (account), 
 			     g_object_ref (protocol));
 
-	/* connect up all signals */ 
 	session_protocol_signals_setup (session, protocol);
 
-	/* clean up */
 	g_object_unref (protocol);
 			
 	return TRUE;
@@ -882,53 +875,19 @@ gossip_session_remove_account (GossipSession *session,
 
 	priv = GET_PRIV (session);
 
-	/* get protocol details for this account */
 	protocol = g_hash_table_lookup (priv->accounts, account);
 	
 	if (!protocol) {
-		/* not added in the first place */
+		/* Not added in the first place. */
 		return TRUE;
 	}
 
-	/* remove from list */
 	priv->protocols = g_list_remove (priv->protocols, protocol);
 	g_object_unref (protocol);
 
-	/* remove from hash table */
 	return g_hash_table_remove (priv->accounts, account);
 }
 
-GossipAccount * 
-gossip_session_find_account (GossipSession *session,
-			     GossipContact *contact)
-{
-	GossipSessionPriv *priv;
-	GossipAccount     *account = NULL;
-	FindAccount       *fa;
-
-	g_return_val_if_fail (GOSSIP_IS_SESSION (session), NULL);
-	g_return_val_if_fail (GOSSIP_IS_CONTACT (contact), NULL);
-	
-	priv = GET_PRIV (session);
-
-	fa = g_new0 (FindAccount, 1);
-	
-	fa->contact_id = g_strdup (gossip_contact_get_id (contact));
-
-	g_hash_table_foreach (priv->accounts, 
-			      (GHFunc)session_find_account_foreach_cb,
-			      fa);
-
-	if (fa->account) {
-		account = fa->account;
-	}
-
-	g_free (fa->contact_id);
-	g_free (fa);
-
-	return account;
-}
-
 static void
 session_find_account_foreach_cb (GossipAccount  *account,
 				 GossipProtocol *protocol,
@@ -939,6 +898,70 @@ session_find_account_foreach_cb (GossipAccount  *account,
 	}
 }
 
+GossipAccount * 
+gossip_session_find_account (GossipSession *session,
+			     GossipContact *contact)
+{
+	GossipSessionPriv *priv;
+	FindAccount        fa;
+
+	g_return_val_if_fail (GOSSIP_IS_SESSION (session), NULL);
+	g_return_val_if_fail (GOSSIP_IS_CONTACT (contact), NULL);
+	
+	priv = GET_PRIV (session);
+
+	fa.contact_id = gossip_contact_get_id (contact);
+	fa.account = NULL;
+
+	g_hash_table_foreach (priv->accounts, 
+			      (GHFunc) session_find_account_foreach_cb,
+			      &fa);
+
+	return fa.account;
+}
+
+static void
+session_find_account_for_own_contact_foreach_cb (GossipAccount  *account,
+						 GossipProtocol *protocol,
+						 FindAccount    *fa)
+{
+	GossipContact *own_contact;
+	const gchar   *id;
+
+	own_contact = gossip_protocol_get_own_contact (protocol);
+	if (!own_contact) {
+		return;
+	}
+	
+	id = gossip_contact_get_id (own_contact);
+	
+	if (g_str_equal (id, fa->contact_id)) {
+		fa->account = g_object_ref (account);
+	}
+}
+
+GossipAccount * 
+gossip_session_find_account_for_own_contact (GossipSession *session,
+					     GossipContact *own_contact)
+{
+	GossipSessionPriv *priv;
+	FindAccount        fa;
+
+	g_return_val_if_fail (GOSSIP_IS_SESSION (session), NULL);
+	g_return_val_if_fail (GOSSIP_IS_CONTACT (own_contact), NULL);
+	
+	priv = GET_PRIV (session);
+
+	fa.contact_id = gossip_contact_get_id (own_contact);
+	fa.account = NULL;
+
+	g_hash_table_foreach (priv->accounts, 
+			      (GHFunc) session_find_account_for_own_contact_foreach_cb,
+			      &fa);
+
+	return fa.account;
+}
+
 static void
 session_account_added_cb (GossipAccountManager *manager,
 			  GossipAccount        *account,
@@ -955,50 +978,6 @@ session_account_removed_cb (GossipAccountManager *manager,
 	gossip_session_remove_account (session, account);
 }
 
-void
-gossip_session_connect (GossipSession *session,
-			GossipAccount *account,
-			gboolean       startup)
-{
-	GossipSessionPriv *priv;
-	ConnectAccounts   *data;
-	
-	g_return_if_fail (GOSSIP_IS_SESSION (session));
-
-	priv = GET_PRIV (session);
-
-	g_return_if_fail (priv->accounts != NULL);
-
-	g_signal_emit (session, signals[CONNECTING], 0);
-
-	/* temporary */
-	if (!priv->presence) {
-		priv->presence = gossip_presence_new_full (GOSSIP_PRESENCE_STATE_AVAILABLE, 
-							   NULL);
-	}
-
-	/* connect one account if provided */
-	if (account) {
-		g_return_if_fail (GOSSIP_IS_ACCOUNT (account));
-		
-		session_connect (session, account);
-		return;
-	}
-
-	/* connect ALL accounts if no one account is specified */
-	data = g_new0 (ConnectAccounts, 1);
-
-	data->session = g_object_ref (session);
-	data->startup = startup;
-
-	g_hash_table_foreach (priv->accounts,
-			      (GHFunc)session_connect_foreach_cb,
-			      data);
-
-	g_object_unref (data->session);
-	g_free (data);
-}
-
 static void
 session_connect (GossipSession *session,
 		 GossipAccount *account) 
@@ -1014,56 +993,110 @@ session_connect (GossipSession *session,
 		return;
 	}
 	
-	g_signal_emit (session, signals[PROTOCOL_CONNECTING], 0, account);
+	g_signal_emit (session, signals[PROTOCOL_CONNECTING], 0, account, protocol);
+	priv->connecting_counter++;
 
-	/* can we not just pass the GossipAccount on the GObject init? */
+	/* Can we not just pass the GossipAccount on the GObject init? */
 	gossip_protocol_setup (protocol, account);
 	
-	/* setup the network connection */
+	/* Setup the network connection */
 	gossip_protocol_login (protocol);
 }
 
 static void
-session_connect_foreach_cb (GossipAccount   *account,
-			    GossipProtocol  *protocol,
-			    ConnectAccounts *data)
+session_connect_match_foreach_cb (GossipAccount  *account,
+				  GossipProtocol *protocol,
+				  ConnectFind    *cf)
 {
-	GossipSessionPriv *priv;
-
-	priv = GET_PRIV (data->session);
-
-	if (data->startup && 
-	    !gossip_account_get_auto_connect (account)) {
+	if (!cf->first && !cf->same) {
 		return;
 	}
 
-	session_connect (data->session, account);
+	if (cf->first) {
+		cf->first = FALSE;
+		cf->last = gossip_account_get_auto_connect (account);
+		return;
+	}
+
+	if (gossip_account_get_auto_connect (account) == cf->last) {
+		cf->same = TRUE;
+	} else {
+		cf->same = FALSE;
+	}
+}
+
+static void
+session_connect_foreach_cb (GossipAccount  *account,
+			    GossipProtocol *protocol,
+			    ConnectData    *cd)
+{
+	/* Connect it if this is a startup request and we have an auto
+	 * connecting account.
+	 */
+	if (cd->startup) {
+		if (gossip_account_get_auto_connect (account)) { 
+			session_connect (cd->session, account);
+		}
+		return; 
+	} 
+
+	/* Connect it if we should be connecting ALL accounts anyway. */
+	if (cd->connect_all) {
+		session_connect (cd->session, account);
+	} 
+	else if (gossip_account_get_auto_connect (account)) {
+		session_connect (cd->session, account);
+	}
 }
 
 void
-gossip_session_disconnect (GossipSession *session,
-			   GossipAccount *account)
+gossip_session_connect (GossipSession *session,
+			GossipAccount *account,
+			gboolean       startup)
 {
 	GossipSessionPriv *priv;
-
+	ConnectFind        cf;
+	ConnectData        cd;
+	
 	g_return_if_fail (GOSSIP_IS_SESSION (session));
 
 	priv = GET_PRIV (session);
 
-	g_signal_emit (session, signals[DISCONNECTING], 0);
+	g_signal_emit (session, signals[CONNECTING], 0);
 
-	/* disconnect one account if provided */
+	/* Temporary */
+	if (!priv->presence) {
+		priv->presence = gossip_presence_new_full (GOSSIP_PRESENCE_STATE_AVAILABLE, 
+							   NULL);
+	}
+
+	/* Connect one account if provided */
 	if (account) {
-		session_disconnect (session, account);
+		session_connect (session, account);
 		return;
 	}
 
-	/* disconnect ALL accounts if no one account is specified */
-	g_hash_table_foreach (priv->accounts,
-			      (GHFunc)session_disconnect_foreach_cb,
-			      session);
-}
+	/* We do some clever stuff here and check that ALL accounts
+	 * are either auto connecting or not, if they are all the same
+	 * then we connect ALL accounts, if not, we just connect the
+	 * auto connect accounts 
+	 */
+	cf.first = TRUE;
+	cf.same = TRUE;
 
+	g_hash_table_foreach (priv->accounts,
+			      (GHFunc) session_connect_match_foreach_cb,
+			      &cf);
+
+	cd.connect_all = cf.same;
+	cd.session = session;
+	cd.startup = startup;
+
+	/* Connect ALL accounts if no one account is specified */
+	g_hash_table_foreach (priv->accounts,
+			      (GHFunc) session_connect_foreach_cb,
+			      &cd);
+}
 
 static void
 session_disconnect (GossipSession *session,
@@ -1078,11 +1111,7 @@ session_disconnect (GossipSession *session,
 	
 	protocol = g_hash_table_lookup (priv->accounts, account);
 	
-/* 	if (!gossip_protocol_is_connected (protocol)) { */
-/* 		return; */
-/* 	} */
-
-	g_signal_emit (session, signals[PROTOCOL_DISCONNECTING], 0, account);
+	g_signal_emit (session, signals[PROTOCOL_DISCONNECTING], 0, account, protocol);
 	
 	gossip_protocol_logout (protocol);
 }
@@ -1099,6 +1128,30 @@ session_disconnect_foreach_cb (GossipAccount  *account,
 	session_disconnect (session, account);
 }
 
+void
+gossip_session_disconnect (GossipSession *session,
+			   GossipAccount *account)
+{
+	GossipSessionPriv *priv;
+
+	g_return_if_fail (GOSSIP_IS_SESSION (session));
+
+	priv = GET_PRIV (session);
+
+	g_signal_emit (session, signals[DISCONNECTING], 0);
+
+	/* Disconnect one account if provided. */
+	if (account) {
+		session_disconnect (session, account);
+		return;
+	}
+
+	/* Disconnect ALL accounts if none is specified. */
+	g_hash_table_foreach (priv->accounts,
+			      (GHFunc)session_disconnect_foreach_cb,
+			      session);
+}
+
 void 
 gossip_session_send_message (GossipSession *session, 
 			     GossipMessage *message)
@@ -1106,39 +1159,24 @@ gossip_session_send_message (GossipSession *session,
 	GossipSessionPriv *priv;
 	GossipContact     *contact;
 	GossipAccount     *account;
-	GList             *l;
-
+	GossipProtocol    *protocol;
+		
 	g_return_if_fail (GOSSIP_IS_SESSION (session));
 	g_return_if_fail (message != NULL);
 
 	priv = GET_PRIV (session);
-	
-	contact = gossip_message_get_recipient (message);
-	account = gossip_session_find_account (session, contact);
-
-	if (account) {
-		GossipProtocol *protocol;
-
-		protocol = g_hash_table_lookup (priv->accounts, account);
-		
-		gossip_protocol_send_message (protocol, message);
-
-		g_object_unref (account);
 
+	contact = gossip_message_get_sender (message);
+	account = gossip_session_find_account_for_own_contact (session, contact);
+	if (!account) {
+		g_warning ("Could not find the account to send message from.");
 		return;
 	}
+	
+	protocol = g_hash_table_lookup (priv->accounts, account);
+	gossip_protocol_send_message (protocol, message);
 
-	/* NOTE: is this right? look for the account based on the
-	   recipient, if not known then send to ALL?? */
-	for (l = priv->protocols; l; l = l->next) {
-		GossipProtocol *protocol = GOSSIP_PROTOCOL (l->data);
-
-		if (!gossip_protocol_is_connected (protocol)) {
-			continue;
-		}
-
-		gossip_protocol_send_message (protocol, message);
-	}
+	g_object_unref (account);
 }
 
 void
@@ -1154,7 +1192,14 @@ gossip_session_send_composing (GossipSession  *session,
 	priv = GET_PRIV (session);
 
 	protocol = session_get_protocol (session, contact);
-	g_return_if_fail (GOSSIP_IS_PROTOCOL (protocol));
+	if (!protocol) {
+		/* We don't warn here because if no protocol is found,
+		 * it is likely that it is because the contact is a
+		 * temporary contact, instead we just silently ignore
+		 * it and don't send composing. 
+		 */
+		return;
+	}
 
 	if (!gossip_protocol_is_connected (protocol)) {
 		return;
@@ -1221,10 +1266,35 @@ gossip_session_is_connected (GossipSession *session,
 		return gossip_protocol_is_connected (protocol);
 	} 
 
-	/* fall back to counter if no account is provided */
+	/* Fall back to counter if no account is provided */
+	gossip_session_count_accounts (session, NULL, NULL, NULL);
+
 	return (priv->connected_counter > 0);
 }
 
+gboolean
+gossip_session_is_connecting (GossipSession *session,
+			      GossipAccount *account)
+{
+	GossipSessionPriv *priv;
+	GossipProtocol    *protocol;
+
+	g_return_val_if_fail (GOSSIP_IS_SESSION (session), FALSE);
+	
+	priv = GET_PRIV (session);
+
+	if (account) {
+		protocol = g_hash_table_lookup (priv->accounts, account);
+		
+		return gossip_protocol_is_connecting (protocol);
+	} 
+
+	/* Fall back to counter if no account is provided */
+	gossip_session_count_accounts (session, NULL, NULL, NULL);
+
+	return (priv->connecting_counter > 0);
+}
+
 const gchar *
 gossip_session_get_active_resource (GossipSession *session,
 				    GossipContact *contact)
@@ -1234,8 +1304,9 @@ gossip_session_get_active_resource (GossipSession *session,
 
 	g_return_val_if_fail (GOSSIP_IS_SESSION (session), NULL);
 	
-	/* get the activate resource, needed to be able to lock the
-	   chat against a certain resource */
+	/* Get the activate resource, needed to be able to lock the
+	 * chat against a certain resource.
+	 */
 
 	priv = GET_PRIV (session);
 
@@ -1278,8 +1349,7 @@ gossip_session_get_ft_provider (GossipSession *session,
 
 	priv = GET_PRIV (session);
 
-	protocol = g_hash_table_lookup (priv->accounts, 
-					gossip_account_get_name (account));
+	protocol = g_hash_table_lookup (priv->accounts, account);
 
 	g_return_val_if_fail (GOSSIP_IS_PROTOCOL (protocol), NULL);
 
@@ -1373,8 +1443,9 @@ gossip_session_remove_contact (GossipSession *session,
 	g_return_if_fail (GOSSIP_IS_SESSION (session));
         g_return_if_fail (GOSSIP_IS_CONTACT (contact));
 	
-	/* get the activate resource, needed to be able to lock the
-	   chat against a certain resource */
+	/* Get the activate resource, needed to be able to lock the
+	 * chat against a certain resource.
+	 */
 
 	priv = GET_PRIV (session);
 
@@ -1493,13 +1564,13 @@ gossip_session_get_own_contact (GossipSession *session,
 	return gossip_protocol_get_own_contact (protocol);
 }
 
-
 GList *
 gossip_session_get_groups (GossipSession *session)
 {
 	GossipSessionPriv *priv;
-	GList             *l, *all_groups = NULL;
-	
+	GList             *l, *j;
+	GList             *all_groups = NULL;
+
 	g_return_val_if_fail (GOSSIP_IS_SESSION (session), NULL);
 
 	priv = GET_PRIV (session);
@@ -1511,8 +1582,12 @@ gossip_session_get_groups (GossipSession *session)
 		protocol = l->data;
 
 		groups = gossip_protocol_get_groups (protocol);
-		if (groups) {
-			all_groups = g_list_concat (all_groups, groups);
+		for (j = groups; j; j = j->next) {
+			if (!g_list_find_custom (all_groups, 
+						 j->data, 
+						 (GCompareFunc) strcmp)) {
+				all_groups = g_list_append (all_groups, j->data);
+			}
 		}
 	}
 
@@ -1707,6 +1782,6 @@ gossip_session_get_version (GossipSession          *session,
 	}
 	
 	return gossip_protocol_get_version (protocol, contact, 
-						  callback, user_data,
-						  error);
+					    callback, user_data,
+					    error);
 }
