@@ -21,9 +21,7 @@
 #ifndef __GOSSIP_APP_H__
 #define __GOSSIP_APP_H__
 
-#include <glib-object.h>
 #include <gtk/gtkwidget.h>
-#include <gconf/gconf-client.h>
 
 #include <libgossip/gossip-account-manager.h>
 #include <libgossip/gossip-chatroom-manager.h>
@@ -40,39 +38,34 @@
 #define GOSSIP_IS_APP_CLASS(k)  (G_TYPE_CHECK_CLASS_TYPE ((k), GOSSIP_TYPE_APP))
 #define GOSSIP_APP_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), GOSSIP_TYPE_APP, GossipAppClass))
 
-
 typedef struct _GossipApp      GossipApp;
 typedef struct _GossipAppClass GossipAppClass;
 typedef struct _GossipAppPriv  GossipAppPriv;
 
-
 struct _GossipApp {
-        GObject        parent;
-
-        GossipAppPriv *priv;
+        GObject parent;
 };
 
-
 struct _GossipAppClass {
         GObjectClass parent_class;
 };
 
-
 GType                  gossip_app_get_type             (void) G_GNUC_CONST;
-
-void                   gossip_app_connect              (gboolean              startup);
+void                   gossip_app_connect              (GossipAccount        *account,
+							gboolean              startup);
 void                   gossip_app_net_down             (void);
 void                   gossip_app_net_up               (void);
-
-void                   gossip_app_create               (GossipAccountManager *manager);
+void                   gossip_app_create               (GossipSession        *session,
+							GossipAccountManager *manager);
 GossipApp *            gossip_app_get                  (void);
-
 gboolean               gossip_app_is_connected         (void);
-void                   gossip_app_force_non_away       (void);
-
+gboolean               gossip_app_is_window_visible    (void);
+void                   gossip_app_toggle_visibility    (void);
+void                   gossip_app_set_visibility       (gboolean              visible);
+void                   gossip_app_set_not_away         (void);
+void                   gossip_app_set_presence         (GossipPresenceState   state,
+							const gchar          *status);
 GtkWidget *            gossip_app_get_window           (void);
-GConfClient *          gossip_app_get_gconf_client     (void);
-
 GossipSession *        gossip_app_get_session          (void);
 GossipChatroomManager *gossip_app_get_chatroom_manager (void);
 GossipChatManager *    gossip_app_get_chat_manager     (void);
