@@ -46,15 +46,25 @@ struct _GossipContactListClass {
 	GtkTreeViewClass       parent_class;
 };
 
-GType               gossip_contact_list_get_type     (void) G_GNUC_CONST;
-GossipContactList * gossip_contact_list_new (void);
+GType              gossip_contact_list_get_type           (void) G_GNUC_CONST;
 
-GossipContact *  gossip_contact_list_get_selected  (GossipContactList *list);
-char *           gossip_contact_list_get_selected_group (GossipContactList *list);
+GossipContactList *gossip_contact_list_new                (void);
+
+GossipContact *    gossip_contact_list_get_selected       (GossipContactList *list);
+char *             gossip_contact_list_get_selected_group (GossipContactList *list);
+
+gboolean           gossip_contact_list_get_show_offline   (GossipContactList *list);
+gboolean           gossip_contact_list_get_show_avatars   (GossipContactList *list);
+
+void               gossip_contact_list_set_show_offline   (GossipContactList *list,
+							   gboolean           show_offline);
+void               gossip_contact_list_set_show_avatars   (GossipContactList *list,
+							   gboolean           show_avatars);
+
+GtkWidget *        gossip_contact_list_get_contact_menu   (GossipContactList *list,
+							   GossipContact     *contact);
+GtkWidget *        gossip_contact_list_get_group_menu     (GossipContactList *list);
 
-gboolean     gossip_contact_list_get_show_offline (GossipContactList *list);
-void         gossip_contact_list_set_show_offline (GossipContactList *list,
-						   gboolean           show_offline);
 
 #endif /* __GOSSIP_CONTACT_LIST_H__ */
 
